// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
// SPDX-FileCopyrightText: 2019 Evan Welsh

/* exported LocaleCategory, bindtextdomain, dcgettext, dgettext, dngettext,
    domain, dpgettext, gettext, ngettext, pgettext, setlocale, textdomain */

var {
    LocaleCategory,
    bindtextdomain,
    dcgettext,
    dgettext,
    dngettext,
    domain,
    dpgettext,
    gettext,
    ngettext,
    pgettext,
    setlocale,
    textdomain,
} = imports._gettext;
