// Sabotage the import of imports.gi.WarnLib!
// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
// SPDX-FileCopyrightText: 2016 Philip Chimento <philip.chimento@gmail.com>

k$^s^%$#^*($%jdghdsfjkgd
