// Sabotage the import of imports.gi.GIMarshallingTests!
// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
// SPDX-FileCopyrightText: 2023 Philip Chimento <philip.chimento@gmail.com>

var _init = {thingsThatThisObjectIsNot: ['callable']};
