// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
// SPDX-FileCopyrightText: 2018 Philip Chimento <philip.chimento@gmail.com>
print('1');
print('2');
print('3');
(function () {
    print('4');
})();
print('5');
